==> dialserver-0.3.3.tar.gz: dialserver-0.3.3/password.c <==
// password.c -- Password file management
//
// Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
//
// This file is part of dialserver.
//
// dialserver is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// dialserver is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// dialserver; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//
#include "common.h"
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif

#ifdef HAVE_SEARCH_H
#include <search.h>	// hash table management
#endif
#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "password.h"
#include "util.h"
#include "conf.h"
 
enum { PS_INITIAL, PS_RELOAD, PS_READY } pass_status = PS_INITIAL;
char *pass_file = NULL;
FILE *pass_cmdwrite, *pass_cmdread;
extern char **environ;

void pass_reloadfile(void)
{
	if (pass_status==PS_READY) 
		pass_status = PS_RELOAD; 
}

// Read the file into a hash table
int pass_readfile(char *filename)
{
	FILE *file;
	char *thisline, *nextline;
	char *username, *password;
	ENTRY entry, *found;
	int rc;
	struct stat statbuf;
	
	// DEBUG: printf ("%s\n",filename);

	// pass_status values: 0: relax  1: just reload 2: do nothing
	if (pass_status==PS_READY) {
		return 0;
	}
	else if (pass_status==PS_INITIAL) {
		// relax
	}
	else { // if (pass_status==PS_RELOAD) 
		hdestroy();		// Deallocate hash table
		free(pass_file);	// Free image of the config file
		pass_file=NULL;
	}
	if (!hcreate(40)) {
		perror("hcreate");
		exit(1);
	}
	pass_status = PS_READY; /* Even if we fail, we are done for now */
	file = fopen(filename,"r");
	if (!file)
		return 0; // oops - no file
	if (fstat(fileno(file),&statbuf)) {
		fclose(file);
		return 0; // oops - maybe it's dev/zero
	}
	pass_file=xmalloc(statbuf.st_size+1); // FIXME (?) breaks for password files >= 2GB!
	pass_file[statbuf.st_size]=0; // Null terminated file - now we can use strchr. Oh, don't put nulls in your password file.
	rc = fread(pass_file,1,statbuf.st_size,file);
	if (ferror(file)) {
		free(pass_file);
		pass_file=NULL;
		fclose(file);
		return 0; // oops
	}
	fclose(file);

	for (thisline = pass_file; thisline; thisline=nextline) {
		nextline=strchr(thisline,'\n');
		if (nextline) {
			*nextline=0; // Don't let processing interfere with next line
			nextline++;
		}
		username = strtok(thisline,":");
		password = strtok(NULL,":\n\r");
		if (username && password) {
			entry.key=username;
			(char *)entry.data=password;
		}
		found = hsearch(entry,ENTER);
		if (found==NULL) {
			return 0;  // oops - out of memory
		}
	}
	return 1; // looks good
}

// Read the file into a hash table
int pass_checkfileauth(char *user, char *password)
{
	ENTRY entry, *found;
	char *crypted;
	pass_readfile(conf.passwdfile);
	entry.key = user;
	entry.data = NULL;
	found = hsearch(entry,FIND);
	if (!found)
		return 0;
	crypted = crypt(password,(char *)found->data);
	if (!crypted)
		return 0;
	if (strcmp(crypted,(char *)found->data)==0)
		return 1;
	return 0;
}

// Open a two-way pipe to the command specified
int pass_cmdreadfile(char *cmdname)
{
	int writeto[2], readfrom[2];
	pid_t pid;
	// filedes[0] is for reading, filedes[1] is for writing

	if ( pass_cmdread && pass_cmdwrite ) return 1;
	if (! conf.cmdcheckpasswd || conf.cmdcheckpasswd[0]==0) {
		return 0;
	}
	
	if (pipe(writeto)==-1 || pipe(readfrom)==-1 ) {
		perror("pipe");
		exit(1);
	}
	pid = fork();
	if (pid == -1) {
		perror("fork");
		exit(1);
	}
	if (pid == 0) {
		char *argv[4];
		int fnum,bad;
		// close all inherited file descriptors so we don't end up with
		// dangling sockets
		dup2(writeto[0],0);  // stdin is written to
		dup2(readfrom[1],1); // stdout is read from
		for (fnum=3, bad=0; fnum < 250 && bad<10 ; fnum++) {
			bad += (close(fnum) == -1);
		}
		argv[0] = "sh";
		argv[1] = "-c";
		argv[2] = cmdname;
		argv[3] = NULL;
		execve("/bin/sh", argv, environ);
		exit(127); // oops
	}
	if ( (pass_cmdwrite=fdopen(writeto[1],"w")) == NULL || 
		(pass_cmdread=fdopen(readfrom[0],"r")) == NULL) {
		perror("fdopen");
		exit(1);
	}
	if ( ! pass_cmdwrite ) return 0;
	if ( ! pass_cmdread ) return 0;
	return 1;
}

// query authentication details against a squid auth program
int pass_cmdcheckauth(char *user, char *password)
{
	size_t rc;
	char result[40];
	char *res;
	if ( ! pass_cmdreadfile(conf.cmdcheckpasswd) ) return 0;
	rc = fprintf(pass_cmdwrite,"%s %s\n",user,password);
	if (rc <= 0) return 0;
	if (fflush(pass_cmdwrite) != 0) return 0;
	// FIXME -- this should use select and timeout on a failure ...
	res = fgets(result,sizeof(result),pass_cmdread);
	if (! res) return 0;
	//debug only
	// fprintf(stderr,"Read %s\n",result);
	return (strncmp(result,"OK",2)==0);
}

// Try a few authorisation methods ...
int pass_checkauth(char *user, char *password)
{
	return pass_checkfileauth(user,password) ||
		pass_cmdcheckauth(user,password);
}

// Read the file from $HOME/.dialserver-details
// If uid==getuid() the file may be written too
int pass_getauthinfo(uid_t uid, char *authinfo, int maxlen, char *prefix, char *suffix)
{
	struct passwd *pwd;
	char authfilename[128]; // HOME=/u/s/r/xxx/.../123/md5sum? => problem
#define AUTHLEN 15 /* random auth string length */
	char authrandom[AUTHLEN+1];
	int result, bytecount;
	int i;
	FILE *authfile = NULL;
	mode_t omask;

	omask = umask(0077);
	pwd = getpwuid( uid );
	if (!pwd) {
		perror("getpwuid");
		result = -1;
		goto end;
	}
	authfilename[sizeof(authfilename)-1] =0;
	snprintf(authfilename,sizeof(authfilename)-1,
		"%s/.dialserver.key",pwd->pw_dir);
	// Read the auth details from the auth file
	authfile = fopen(authfilename,"a+");
	if (!authfile) {
		perror(authfilename);
		result = -1;
		goto end;
	}
	rewind(authfile); // hey what error?
	if (fread(authrandom,1,AUTHLEN,authfile)<AUTHLEN) {
		if (getuid() != uid) {
			result = -1;
			goto end;
		}
		// If it's not there, generate a random string
		rewind(authfile);
		srandom((int)time(NULL));  // hack me hack me
		for (i=0; i<AUTHLEN; i++) {
			authrandom[i] = (char) (' ' + 1 + (random() & 63 ));
		}
		bytecount = fwrite(authrandom,1,AUTHLEN,authfile);
		if (bytecount<AUTHLEN) {
			perror(authfilename);
			result = -1;
			goto end;
		}
		fputc('\n',authfile); // decoration
	}
	else {
		for (i=0; i<AUTHLEN; i++) {
			// die if file contains control chars
			if (authrandom[i] <= ' ') {
				result = -1;
				goto end;
			}
		}
	}
	authrandom[AUTHLEN] = 0;
	authinfo[maxlen-1] = 0;
	snprintf(authinfo,maxlen-1,"%s%d %s%s",
		prefix,(int)uid,authrandom,suffix);
	result = 0;
end:
	if (authfile) {
		fclose(authfile);
	}
	umask(omask);
	return result;
}

==> dialserver-0.3.8.tar.gz: dialserver-0.3.8/password.c <==
// password.c -- Password file management
//
// Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
//
// This file is part of dialserver.
//
// dialserver is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// dialserver is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// dialserver; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//
#include "common.h"
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif

#ifdef HAVE_SEARCH_H
#include <search.h>	// hash table management
#endif
#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "password.h"
#include "util.h"
#include "conf.h"
 
enum { PS_INITIAL, PS_RELOAD, PS_READY } pass_status = PS_INITIAL;
char *pass_file = NULL;
FILE *pass_cmdwrite, *pass_cmdread;
extern char **environ;

void pass_reloadfile(void)
{
	if (pass_status==PS_READY) 
		pass_status = PS_RELOAD; 
}

// Read the file into a hash table
int pass_readfile(char *filename)
{
	FILE *file;
	char *thisline, *nextline;
	char *username, *password;
	ENTRY entry, *found;
	int rc;
	struct stat statbuf;
	
	// DEBUG: printf ("%s\n",filename);

	// pass_status values: 0: relax  1: just reload 2: do nothing
	if (pass_status==PS_READY) {
		return 0;
	}
	else if (pass_status==PS_INITIAL) {
		// relax
	}
	else { // if (pass_status==PS_RELOAD) 
		hdestroy();		// Deallocate hash table
		free(pass_file);	// Free image of the config file
		pass_file=NULL;
	}
	if (!hcreate(40)) {
		perror("hcreate");
		exit(1);
	}
	pass_status = PS_READY; /* Even if we fail, we are done for now */
	file = fopen(filename,"r");
	if (!file)
		return 0; // oops - no file
	if (fstat(fileno(file),&statbuf)) {
		fclose(file);
		return 0; // oops - maybe it's dev/zero
	}
	pass_file=xmalloc(statbuf.st_size+1); // FIXME (?) breaks for password files >= 2GB!
	pass_file[statbuf.st_size]=0; // Null terminated file - now we can use strchr. Oh, don't put nulls in your password file.
	rc = fread(pass_file,1,statbuf.st_size,file);
	if (ferror(file)) {
		free(pass_file);
		pass_file=NULL;
		fclose(file);
		return 0; // oops
	}
	fclose(file);

	for (thisline = pass_file; thisline; thisline=nextline) {
		nextline=strchr(thisline,'\n');
		if (nextline) {
			*nextline=0; // Don't let processing interfere with next line
			nextline++;
		}
		username = strtok(thisline,":");
		password = strtok(NULL,":\n\r");
		if (username && password) {
			entry.key=username;
			entry.data=(void *)password;
		}
		else {
			return 0; // oops - no info to work with
		}
		found = hsearch(entry,ENTER);
		if (found==NULL) {
			return 0;  // oops - out of memory
		}
	}
	return 1; // looks good
}

// Read the file into a hash table
int pass_checkfileauth(char *user, char *password)
{
	ENTRY entry, *found;
	char *crypted;
	pass_readfile(conf.passwdfile);
	entry.key = user;
	entry.data = NULL;
	found = hsearch(entry,FIND);
	if (!found)
		return 0;
	crypted = crypt(password,(char *)found->data);
	if (!crypted)
		return 0;
	if (strcmp(crypted,(char *)found->data)==0)
		return 1;
	return 0;
}

// Open a two-way pipe to the command specified
int pass_cmdreadfile(char *cmdname)
{
	int writeto[2], readfrom[2];
	pid_t pid;
	// filedes[0] is for reading, filedes[1] is for writing

	if ( pass_cmdread && pass_cmdwrite ) return 1;
	if (! conf.cmdcheckpasswd || conf.cmdcheckpasswd[0]==0) {
		return 0;
	}
	
	if (pipe(writeto)==-1 || pipe(readfrom)==-1 ) {
		perror("pipe");
		exit(1);
	}
	pid = fork();
	if (pid == -1) {
		perror("fork");
		exit(1);
	}
	if (pid == 0) {
		char *argv[4];
		int fnum,bad;
		// close all inherited file descriptors so we don't end up with
		// dangling sockets
		dup2(writeto[0],0);  // stdin is written to
		dup2(readfrom[1],1); // stdout is read from
		for (fnum=3, bad=0; fnum < 250 && bad<10 ; fnum++) {
			bad += (close(fnum) == -1);
		}
		argv[0] = "sh";
		argv[1] = "-c";
		argv[2] = cmdname;
		argv[3] = NULL;
		execve("/bin/sh", argv, environ);
		exit(127); // oops
	}
	if ( (pass_cmdwrite=fdopen(writeto[1],"w")) == NULL || 
		(pass_cmdread=fdopen(readfrom[0],"r")) == NULL) {
		perror("fdopen");
		exit(1);
	}
	if ( ! pass_cmdwrite ) return 0;
	if ( ! pass_cmdread ) return 0;
	return 1;
}

// query authentication details against a squid auth program
int pass_cmdcheckauth(char *user, char *password)
{
	size_t rc;
	char result[40];
	char *res;
	fd_set exceptions, readable, writable;

	struct timeval timeout;
	int nfds;

	if (fflush(pass_cmdread) != 0) return 0;
	if ( ! pass_cmdreadfile(conf.cmdcheckpasswd) ) return 0;
	rc = fprintf(pass_cmdwrite,"%s %s\n",user,password);
	if (rc <= 0) return 0;
	if (fflush(pass_cmdwrite) != 0) return 0;
	// FIXED -- this should use select and timeout on a failure ...
	//fprintf(stderr,"Reading from fd %d\n",fileno(pass_cmdread));

	// Check that fd is readable
	FD_ZERO(&readable);
	FD_ZERO(&writable);
	FD_ZERO(&exceptions);  // Methinks exceptions is only for reading out-of-band stuff
	FD_SET(fileno(pass_cmdread), &readable);
	timeout.tv_sec = 0;
	timeout.tv_usec = 10000; // 0.01 seconds
	nfds = select(FD_SETSIZE, &readable, &writable, &exceptions, &timeout);

	if (nfds <= 0) {
		pass_cmdread = NULL;
		pass_cmdwrite = NULL;
		return 0;
	};

	// Hope for the best here .. skip the select()
	//fprintf(stderr,"Reading from fd %d nfds=%d\n",fileno(pass_cmdread),nfds);
	res = fgets(result,sizeof(result),pass_cmdread);
	//fprintf(stderr,"Reading from fd %d returned %s\n",fileno(pass_cmdread),res);
	if (! res) return 0;
	//debug only
	// fprintf(stderr,"Read %s\n",result);
	return (strncmp(result,"OK",2)==0);
}

// Try a few authorisation methods ...
int pass_checkauth(char *user, char *password)
{
	return pass_checkfileauth(user,password) ||
		pass_cmdcheckauth(user,password);
}

// Read the file from $HOME/.dialserver-details
// If uid==getuid() the file may be written too
int pass_getauthinfo(uid_t uid, char *authinfo, int maxlen, char *prefix, char *suffix)
{
	struct passwd *pwd;
	char authfilename[128]; // HOME=/u/s/r/xxx/.../123/md5sum? => problem
#define AUTHLEN 15 /* random auth string length */
	char authrandom[AUTHLEN+1];
	int result, bytecount;
	int i;
	FILE *authfile = NULL;
	mode_t omask;

	omask = umask(0077);
	pwd = getpwuid( uid );
	if (!pwd) {
		perror("getpwuid");
		result = -1;
		goto end;
	}
	authfilename[sizeof(authfilename)-1] =0;
	snprintf(authfilename,sizeof(authfilename)-1,
		"%s/.dialserver.key",pwd->pw_dir);
	// Read the auth details from the auth file
	authfile = fopen(authfilename,"a+");
	if (!authfile) {
		perror(authfilename);
		result = -1;
		goto end;
	}
	rewind(authfile); // hey what error?
	if (fread(authrandom,1,AUTHLEN,authfile)<AUTHLEN) {
		if (getuid() != uid) {
			result = -1;
			goto end;
		}
		// If it's not there, generate a random string
		rewind(authfile);
		srandom((int)time(NULL));  // hack me hack me
		for (i=0; i<AUTHLEN; i++) {
			authrandom[i] = (char) (' ' + 1 + (random() & 63 ));
		}
		bytecount = fwrite(authrandom,1,AUTHLEN,authfile);
		if (bytecount<AUTHLEN) {
			perror(authfilename);
			result = -1;
			goto end;
		}
		fputc('\n',authfile); // decoration
	}
	else {
		for (i=0; i<AUTHLEN; i++) {
			// die if file contains control chars
			if (authrandom[i] <= ' ') {
				result = -1;
				goto end;
			}
		}
	}
	authrandom[AUTHLEN] = 0;
	authinfo[maxlen-1] = 0;
	snprintf(authinfo,maxlen-1,"%s%d %s%s",
		prefix,(int)uid,authrandom,suffix);
	result = 0;
end:
	if (authfile) {
		fclose(authfile);
	}
	umask(omask);
	return result;
}
